import BaseAnnotation from '../annotation/base'
import { onClickOutside } from '../ui_utils'
import { createSvg } from '../annotation/utils'
import { getActualPoint } from '../annotation/utils'

class TextAnnotation extends BaseAnnotation {
  constructor ({
    container,
    annotation,
    page,
    viewport,
    scale,
    eventBus,
    layer,
    show = false
  }) {
    super({
      container,
      annotation,
      page,
      viewport,
      scale,
      eventBus,
      show
    })
    this.layer = layer
    this.hidden = true
    this.markupContainer = null
    this.outerLineContainer = null

    this.onHandleClick = this.handleClick.bind(this)
    this.onDelete = this.handleDelete.bind(this)

    this.render()
  }

  render () {
    const annotation = this.annotation

    const { start, end } = this.getActualRect(
      this.viewport,
      this.scale
    )
    const rect = this.calculate(start, end)

    const quadPoints = annotation.quadPoints || annotation.coords
    if (!quadPoints) return
    const arr = quadPoints.split(","); // 将字符串按逗号分隔生成数组

    let tempArray = []
    let chunk = []
    for (let i = 0; i < arr.length; i++) {
      chunk.push(arr[i])
      if (chunk.length === 8 || i === arr.length - 1) {
        tempArray.push(chunk)
        chunk = []
      }
    }
    const positionArray = this.getActual(tempArray)

    if (!annotation.borderColor) {
      annotation.borderColor = annotation.color
    }

    if (annotation.type === 'highlight') {
      this.renderHighlight(positionArray)
    } else if (annotation.type === 'underline') {
      this.renderUnderline(positionArray)
    } else if (annotation.type === 'squiggly') {
      this.renderSquiggly(positionArray)
    } else if (annotation.type === 'strikeout') {
      this.renderStrikeout(positionArray)
    }
    this.markupContainer.addEventListener('click', this.onHandleClick)
    
    this.outerLineContainer = document.createElement('div')
    this.outerLineContainer.className = 'outline-container'

    const rectContainer = createSvg(
      "svg"
    )
    rectContainer.style.left = (rect.left - 6) + 'px'
    rectContainer.style.top = (rect.top - 6) + 'px'
    rectContainer.style.width = (rect.width + 12) + 'px'
    rectContainer.style.height = (rect.height + 12) + 'px'

    const recttangle = createSvg(
      "rect", {
        x: 3,
        y: 3,
        width: rect.width + 6,
        height: rect.height + 6,
        fill: '#ffffff',
        'fill-opacity': 0,
        'stroke-width': 1,
        stroke: '#1460F3'
      }
    )
    rectContainer.append(recttangle)

    const deletetButton = createSvg(
      "svg",
      {
        width: "30",
        height: "30",
        viewBox: "0 0 30 30",
        fill: "none",
        class: 'delete-button'
      },
      {
        width: '30px',
        height: '30px'
      }
    )
    deletetButton.addEventListener('click', this.onDelete)
    deletetButton.innerHTML = this.deleteSvgStr

    const left = (rect.left + rect.width - 30) + 'px'
    const top = (rect.top - 38) + 'px'
    deletetButton.style.left = left
    deletetButton.style.top = top

    this.deletetButton = deletetButton
    this.outerLineContainer.append(rectContainer)
    this.outerLineContainer.append(deletetButton)
  }

  getActualRect (viewport, s) {
    const annotation = this.annotation
    const [x1, y1, x2, y2] = annotation.rect.split(',')

    const start = getActualPoint(
      {
        x: x1,
        y: y1
      },
      viewport,
      s
    )
    const end = getActualPoint(
      {
        x: x2,
        y: y2
      },
      viewport,
      s
    )
    return {
      start,
      end
    }
  }

  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  }

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });
    const actualBdWidth = 2

    return {
      top: initRect.top - actualBdWidth,
      left: initRect.left - actualBdWidth,
      width: initRect.width + actualBdWidth * 2,
      height: initRect.height + actualBdWidth * 2,
    }
  }

  handleClick () {
    if (!this.hidden) return
    this.hidden = false
    this.updateTool()
    onClickOutside([this.markupContainer, this.outerLineContainer, this.deletetButton], this.handleOutside.bind(this))
  }

  handleOutside () {
    this.hidden = !this.hidden
    if (this.layer.selectedElementName === this.annotation.name) {
      this.layer.selectedElementName = null
    }
    this.outerLineContainer.remove()
  }

  updateTool () {
    if (this.hidden) {
      if (this.layer.selectedElementName === this.annotation.name) {
        this.layer.selectedElementName = null
      }
      this.outerLineContainer.remove()
    } else {
      if (this.layer.selectedElementName !== this.annotation.name) {
        this.layer.selectedElementName = this.annotation.name
      }
      this.container.append(this.outerLineContainer)
    }
  }

  handleDelete (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    this.handleOutside()
    this.markupContainer.remove()
    this.annotation.isDelete = true
    this.eventBus.dispatch('annotationChange', {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        page: this.page
      }
    })
  }

  getActual (tempArray) {
    const { height, scale } = this.viewport
    const position = tempArray.map((ele) => ({
      left: Math.min(parseInt(ele[0] * scale, 10), parseInt(ele[2] * scale, 10)),
      bottom: Math.max(parseInt(height - ele[5] * scale, 10), parseInt(height - ele[1] * scale, 10)),
      right: Math.max(parseInt(ele[0] * scale, 10), parseInt(ele[2] * scale, 10)),
      top: Math.min(parseInt(height - ele[5] * scale, 10), parseInt(height - ele[1] * scale, 10)),
    }))
    return position
  }
  
  setCss (ele, cssText) {
    if (!ele) return
    if (cssText) {
      for (let key in cssText) {
        ele.style[key] = cssText[key]
      }
    }
  }

  renderHighlight (positionArray) {
    const annotation = this.annotation
    const markupContainer = document.createElement('div')
    markupContainer.className = 'markup'
    this.setCss(markupContainer, {
      mixBlendMode: 'multiply',
      cursor: 'pointer',
      pointerEvents: 'auto'
    })
    for (let i = 0; i < positionArray.length; i++) {
      const currentPositionArray = positionArray[i]
      const container = document.createElement('div')
      this.setCss(container, {
        position: 'absolute',
        opacity: 0.5,
        top: currentPositionArray.top + 'px',
        left: currentPositionArray.left + 'px',
        width: (currentPositionArray.right - currentPositionArray.left) + 'px',
        height: (currentPositionArray.bottom - currentPositionArray.top) + 'px',
        backgroundColor: annotation.borderColor
      })
      markupContainer.append(container)
    }
    this.container.append(markupContainer)
    this.markupContainer = markupContainer
  }

  renderUnderline (positionArray) {
    const annotation = this.annotation
    const markupContainer = document.createElement('div')
    markupContainer.className = 'markup'
    this.setCss(markupContainer, {
      cursor: 'pointer',
      pointerEvents: 'auto'
    })
    for (let i = 0; i < positionArray.length; i++) {
      const currentPositionArray = positionArray[i]
      const container = document.createElement('div')
      this.setCss(container, {
        position: 'absolute',
        mixBlendMode: 'multiply',
        opacity: annotation.transparency || annotation.opacity,
        top: currentPositionArray.top + 'px',
        left: currentPositionArray.left + 'px',
        width: (currentPositionArray.right - currentPositionArray.left) + 'px',
        height: (currentPositionArray.bottom - currentPositionArray.top) + 'px',
        borderBottom: '2px solid ' + annotation.borderColor
      })
      markupContainer.append(container)
    }
    this.container.append(markupContainer)
    this.markupContainer = markupContainer
  }

  renderSquiggly (positionArray) {
    const annotation = this.annotation
    const markupContainer = document.createElement('div')
    markupContainer.className = 'markup'
    this.setCss(markupContainer, {
      cursor: 'pointer',
      pointerEvents: 'auto'
    })
    for (let i = 0; i < positionArray.length; i++) {
      const currentPositionArray = positionArray[i]
      const container = document.createElement('div')
      this.setCss(container, {
        position: 'absolute',
        mixBlendMode: 'multiply',
        overflow: 'hidden',
        opacity: annotation.transparency || annotation.opacity,
        top: currentPositionArray.top + 'px',
        left: currentPositionArray.left + 'px',
        width: (currentPositionArray.right - currentPositionArray.left) + 'px',
        height: (currentPositionArray.bottom - currentPositionArray.top) + 'px',
      })
      
      const squigglyBefore = document.createElement('div')
      this.setCss(squigglyBefore, {
        position: 'absolute',
        width: '100%',
        height: '5px',
        left: '0px',
        bottom: '1px',
        background: `radial-gradient(ellipse, transparent, transparent 8px, ${annotation.borderColor} 9px, ${annotation.borderColor} 10px, transparent 11px)`,
        backgroundSize: '22px 26px',
        backgroundRepeat: 'repeat-x'
      })
      const squigglyAfter = document.createElement('div')
      this.setCss(squigglyAfter, {
        position: 'absolute',
        width: '100%',
        height: '5px',
        left: '11px',
        bottom: '-2px',
        background: `radial-gradient(ellipse, transparent, transparent 8px, ${annotation.borderColor} 9px, ${annotation.borderColor} 10px, transparent 11px)`,
        backgroundSize: '22px 26px',
        backgroundPosition: '0px -22px',
        backgroundRepeat: 'repeat-x'
      })
      container.append(squigglyBefore)
      container.append(squigglyAfter)
      markupContainer.append(container)
    }
    this.container.append(markupContainer)
    this.markupContainer = markupContainer
  }

  renderStrikeout (positionArray) {
    const annotation = this.annotation
    const markupContainer = document.createElement('div')
    markupContainer.className = 'markup'
    this.setCss(markupContainer, {
      cursor: 'pointer',
      pointerEvents: 'auto'
    })
    for (let i = 0; i < positionArray.length; i++) {
      const currentPositionArray = positionArray[i]
      const container = document.createElement('div')
      this.setCss(container, {
        position: 'absolute',
        mixBlendMode: 'multiply',
        opacity: annotation.transparency || annotation.opacity,
        top: currentPositionArray.top + 'px',
        left: currentPositionArray.left + 'px',
        width: (currentPositionArray.right - currentPositionArray.left) + 'px',
        height: (currentPositionArray.bottom - currentPositionArray.top) + 'px'
      })
      const strikeout = document.createElement('div')
      this.setCss(strikeout, {
        position: 'absolute',
        opacity: 0.8,
        top: '50%',
        left: 0,
        width: '100%',
        height: '2px',
        transform: 'translateY(-50%)',
        backgroundColor: annotation.borderColor
      })
      container.append(strikeout)
      markupContainer.append(container)
    }
    this.container.append(markupContainer)
    this.markupContainer = markupContainer
  }
}

export default TextAnnotation
