import { defineStore } from 'pinia'
import core from '@/core'

const type = ['ink', 'line', 'square', 'circle', 'freetext', 'text', 'highlight', 'underline', 'squiggly', 'strikeout', 'image', 'stamp']

export const useDocumentStore = defineStore({
  id: 'document',
  state: () => ({
    totalPages: 0,
    loadingProgress: 0,
    annotationsCount: 0,
    info: null,
    measure: {
      precision: 1,
      scale: {
        origin: {
          num: 1,
          unit: 'pt'
        },
        actual: {
          num: 1,
          unit: 'pt'
        }
      },
      distance: 0
    },
    optionUrl: null,
    markupTool: 'highlight',
    highlight: {
      color: '#FFEC66'
    },
    underline: {
      color: '#FF0000'
    },
    squiggly: {
      color: '#FF0000'
    },
    strikeout: {
      color: '#FF0000'
    },
    shapeTool: 'square',
    square: {
      color: '#FF0000',
      opacity: '0.8',
      stroke: '1pt'
    },
    circle: {
      color: '#FF0000',
      opacity: '0.8',
      stroke: '1pt'
    },
    arrow: {
      color: '#FF0000',
      opacity: '0.8',
      stroke: '1pt'
    },
    line: {
      color: '#FF0000',
      opacity: '0.8',
      stroke: '1pt'
    },
    ink: {
      color: '#FF0000',
      opacity: '0.8',
      stroke: '2pt'
    },
    freetext: {
      color: '#000000'
    },
    text: {
      color: '#000000'
    },
    activeTool: '',
    annotations: null,
    textfield: {
      color: '#DDE9FF'
    },
    checkbox: {
      color: '#DDE9FF',
      borderColor: '#43474D'
    },
    radiobutton: {
      color: '#DDE9FF',
      borderColor: '#43474D'
    },
    listbox: {
      color: '#DDE9FF',
      borderColor: '#43474D'
    },
    combobox: {
      color: '#DDE9FF',
      borderColor: '#43474D'
    },
    pushbutton: {
      color: '#DDE9FF',
      borderColor: '#43474D'
    },
    link: {
      color: '#DDE9FF'
    },
    propertyPanel: {
      fieldName: '',
      isHidden: '0',
      backgroundColor: '#DDE9FF',
      textColor: '#000000',
      fontName: 'Helvetica',
      fontSize: '14',
      alignment: '0',
      value: '',
      multiline: false,
      project: '',
      projectList: '',
      style: '0',
      select: '',
      actionType: '0',
      text: 'OK',
      url: '',
      destPage: '',
      options: []
    },
    compareFiles: {
      oldFile: null,
      oldFilePassword: '',
      newFile: null,
      newFilePassword: ''
    },
    compareSettings: {
      oldFileColor: '#FBBDBF',
      newFileColor: '#93B9FD'
    },
    currentPdfData: null,
    fileHasPwd: false
  }),
  getters: {
    getTotalPages () {
      return this.totalPages
    },
    getLoadingProgress () {
      return this.loadingProgress
    },
    getAnnotationsCount () {
      return this.annotationsCount + ' annotation' + (this.annotationsCount > 1 ? 's' : '')
    },
    getMeasure () {
      return this.measure
    },
    getPrecision () {
      return this.measure.precision === 1 ? 0.1 : 0.01
    },
    getInfo () {
      return this.info
    },
    getOptionUrl () {
      return this.optionUrl
    },
    getMarkupToolState () {
      return this.markupTool
    },
    getLineType () {
      return this.line.type
    },
    getActiveToolColor () {
      if (this.activeTool) {
        return this[this.activeTool].color
      }
    },
    getActiveTool () {
      return this.activeTool
    },
    getShapeToolState () {
      return this.shapeTool
    },
    getAllAnnotations () {
      const annotationsContainers = {
        annotations: {},
        annotationsCount: 0
      }
      const annotations = this.annotations
      for (const page in annotations) {
        const pageAnnotations = annotations[page]
        const  len = pageAnnotations.length
        if (len) {
          for (let i = 0; i < len; i++) {
           // if (pageAnnotations[i].type === 'line' && pageAnnotations[i].tail === 'OpenArrow') continue
            if (pageAnnotations[i] && pageAnnotations[i].type === 'line') {
              let supportLine=true
              if(pageAnnotations[i].tail || pageAnnotations[i].head)
              {
                if (pageAnnotations[i].tail === 'None' && pageAnnotations[i].head === 'None') {
                  supportLine=true
                }
                else
                {
                  if (pageAnnotations[i].head && pageAnnotations[i].head !== 'OpenArrow' && pageAnnotations[i].head !== 'None') {
                    supportLine=false
                    continue
                  }

                  if (pageAnnotations[i].tail && pageAnnotations[i].tail !== 'OpenArrow' && pageAnnotations[i].tail !== 'None') {
                    supportLine=false
                    continue
                  }
                } 
              }
            }

            if (type.includes(pageAnnotations[i].type)) {
              if (!annotationsContainers.annotations[page]) {
                annotationsContainers.annotations[page] = {
                  pageAnnotationsCount: 0,
                  annotations: []
                }
              }
              annotationsContainers.annotations[page].pageAnnotationsCount++
              annotationsContainers.annotationsCount++
              annotationsContainers.annotations[page].annotations.push(pageAnnotations[i])
            }
          }
        }
      }
      return annotationsContainers
    },
    getAnnotations (pageNum) {
      return this.annotations && this.annotations[pageNum]
    },
    getCompareFile (state) {
      return (key) => state.compareFiles[key]
    },
    getCompareFiles () {
      return this.compareFiles
    },
    getCompareSetting (state) {
      return (prop) => state.compareSettings[prop]
    },
    getCurrentPdfData () {
      return this.currentPdfData
    },
    getFilehasPwd () {
      return this.fileHasPwd
    }
  },
  actions: {
    resetSetting () {
      this.totalPages = 0
      this.loadingProgress = 0
      this.annotationsCount = 0
      this.annotations = null
    },
    setTotalPages (totalPages) {
      this.totalPages = totalPages
    },
    setLoadingProgress (percent) {
      this.loadingProgress = percent
    },
    setAnnotationsCount (count) {
      this.annotationsCount = count
    },
    setOptionUrl (optionUrl) {
      this.optionUrl = optionUrl
    },
    setToken (info) {
      this.info = info
    },
    setDistance (distance) {
      this.measure.distance = Math.round(distance * 10) / 10
    },
    setMarkupToolState (tool) {
      this.markupTool = tool
    },
    setMarkupToolColor (color) {
      this[this.markupTool] = color
    },
    setShapeToolState (tool) {
      this.shapeTool = tool
    },
    setLineType (type) {
      this.line.type = type
    },
    setLineColor (color) {
      this.line.color = color
    },
    setActiveTool (tool) {
      this.activeTool = tool
    },
    setActiveToolColor (color) {
      if (this.activeTool) {
        this[this.activeTool].color = color
        core.setTool({
          tool: this.activeTool,
          color: color
        })
      }
    },
    setToolState (tool) {
      if (this.activeTool === tool || !tool) {
        this.activeTool = ''
        core.setTool({
          tool: '',
          color: ''
        })
      } else {
        this.activeTool = tool
        core.setTool({
          tool,
          color: this[tool]?.color || ''
        })
        if (this.activeTool === 'textfield' || this.activeTool === 'checkbox' || this.activeTool === 'radiobutton' || this.activeTool === 'pushbutton') {
          core.setInitProperty()
        }
      }
    },
    initAnnotations (annotations) {
      this.annotations = null
      this.annotations = annotations
    },
    setAnnotations (annotation, type) {
      const annotations = this.annotations
      if (type === 'add') {
        if (!annotations[annotation[page]]) {
          annotations[page] = []
        }
        annotations[page].push(annotation)
      } else {
        const annotationsPage = annotations[annotation[page]]
        const deleteIndex = annotationsPage.findIndex((item) => item.index ===  annotation.index)
        if (type === 'delete') {
          annotationsPage[deleteIndex].isDelete = true
        } else {
          annotationsPage[deleteIndex] = {
            ...annotationsPage[deleteIndex],
            ...annotation
          }
        }
      }
    },
    setPropertyPanel (props) {
      let updatedProps = {}
      for (const item in props) {
        for (const key in this.propertyPanel) {
          if (item === key) {
            updatedProps[key] = props[item]
          }
        }
      }
      core.setPropertyPanel(updatedProps)
    },
    setCompareFiles (key, val) {
      this.compareFiles[key] = val
    },
    initCompareFiles () {
      this.compareFiles.oldFile = null
      this.compareFiles.oldFilePassword = ''
      this.compareFiles.newFile = null
      this.compareFiles.newFilePassword = ''
    },
    setCompareSetting (key, value) {
      this.compareSettings[key] = value
    },
    initCompareSetting () {
      this.compareSettings = {
        oldFileColor: '#FBBDBF',
        newFileColor: '#93B9FD'
      }
    },
    setCurrentPdfData (data) {
      this.currentPdfData = data
    },
    setFileHasPwd (bool) {
      this.fileHasPwd = bool
    }
  }
})
